from queue import Empty

from remoulade import QueueJoinTimeout
from remoulade.common import current_millis


def iter_queue(queue):
    """Iterate over the messages on a queue until it's empty.  Does
    not block while waiting for messages.

    Parameters:
      queue(Queue): The queue to iterate over.

    Returns:
      generator[object]: The iterator.
    """
    while True:
        try:
            yield queue.get_nowait()
        except Empty:
            break


def join_queue(queue, timeout=None):
    """The join() method of standard queues in Python doesn't support
    timeouts.  This implements the same functionality as that method,
    with optional timeout support, by depending the internals of
    Queue.

    Raises:
      QueueJoinTimeout: When the timeout is reached.

    Parameters:
      timeout(Optional[float])
    """
    with queue.all_tasks_done:
        while queue.unfinished_tasks:
            finished_in_time = queue.all_tasks_done.wait(timeout=timeout)
            if not finished_in_time:
                raise QueueJoinTimeout("timed out after %.02f seconds" % timeout)


def join_all(joinables, timeout):
    """Wait on a list of objects that can be joined with a total
    timeout represented by ``timeout``.

    Parameters:
      joinables(object): Objects with a join method.
      timeout(int): The total timeout in milliseconds.
    """
    started, elapsed = current_millis(), 0
    for ob in joinables:
        ob.join(timeout=timeout / 1000)
        elapsed = current_millis() - started
        timeout = max(0, timeout - elapsed)


def q_name(queue_name):
    """Returns the canonical queue name for a given queue."""
    if queue_name.endswith(".DQ") or queue_name.endswith(".XQ"):
        return queue_name[:-3]
    return queue_name


def dq_name(queue_name):
    """Returns the delayed queue name for a given queue.  If the given
    queue name already belongs to a delayed queue, then it is returned
    unchanged.
    """
    if queue_name.endswith(".DQ"):
        return queue_name

    if queue_name.endswith(".XQ"):
        queue_name = queue_name[:-3]
    return queue_name + ".DQ"


def xq_name(queue_name):
    """Returns the dead letter queue name for a given queue.  If the
    given queue name belongs to a delayed queue, the dead letter queue
    name for the original queue is generated.
    """
    if queue_name.endswith(".XQ"):
        return queue_name

    if queue_name.endswith(".DQ"):
        queue_name = queue_name[:-3]
    return queue_name + ".XQ"



from functools import update_wrapper

import pytz
from flask import Blueprint
from flask_apispec import doc, marshal_with
from marshmallow import Schema, ValidationError, fields, validates_schema
from marshmallow.validate import OneOf

import remoulade
from remoulade import get_scheduler
from remoulade.errors import NoScheduler
from remoulade.scheduler import ScheduledJob

from .apispec import validate_schema

scheduler_bp = Blueprint("scheduler", __name__, url_prefix="/scheduled")


class ScheduledJobSchema(Schema):
    """
    Class to validate scheduled jobs to add
    """

    actor_name = fields.Str()
    args = fields.List(fields.Raw(), allow_none=True)
    kwargs = fields.Dict(allow_none=True)
    interval = fields.Int(allow_none=True)
    daily_time = fields.Time(allow_none=True)
    iso_weekday = fields.Int(allow_none=True)
    enabled = fields.Bool(allow_none=True)
    last_queued = fields.DateTime(allow_none=True)
    tz = fields.Str(allow_none=True, validate=OneOf(pytz.all_timezones))

    @validates_schema
    def validate_sort(self, data, **kwargs):
        if data.get("daily_time") is not None and data.get("interval") != 86400:
            raise ValidationError("daily_time can only be used with 24h interval")

    @validates_schema
    def validate_actor_name(self, data, **kwargs):
        actor_name = data.get("actor_name")
        if actor_name not in remoulade.get_broker().actors.keys():
            raise ValidationError(f"No actor named {actor_name} exists")

    @validates_schema
    def validate_tz(self, data, **kwargs):
        daily_time = data.get("daily_time")
        last_queued = data.get("last_queued")
        if (daily_time is not None and daily_time.tzinfo is not None) or (
            last_queued is not None and last_queued.tzinfo is not None
        ):
            raise ValidationError("Datetimes should be naive")


class JobWithHashSchema(ScheduledJobSchema):
    hash = fields.Str()


class ScheduleResponseSchema(Schema):
    result = fields.List(fields.Nested(JobWithHashSchema), allow_none=True)
    error = fields.Str(allow_none=True)


class ScheduledJobsSchema(Schema):
    """
    Class to validate multiple scheduled jobs at once
    """

    jobs = fields.Dict(keys=fields.Str(), values=fields.Nested(ScheduledJobSchema))


def with_scheduler(func):
    def wrapper(*args, **kwargs):
        try:
            scheduler = get_scheduler()
            func(scheduler, *args, **kwargs)
            scheduled_jobs = scheduler.get_redis_schedule()
            return {"result": [job.as_dict() for job in scheduled_jobs.values()]}

        except NoScheduler:
            return {"error": "this route needs a scheduler, but no scheduler was found."}, 400

    return update_wrapper(wrapper, func)


@scheduler_bp.route("/jobs")
@doc(tags=["scheduler"])
@marshal_with(ScheduleResponseSchema)
@with_scheduler
def get_jobs(scheduler):
    pass


@scheduler_bp.route("/jobs", methods=["POST"])
@doc(tags=["scheduler"])
@marshal_with(ScheduleResponseSchema)
@validate_schema(ScheduledJobSchema)
@with_scheduler
def add_job(scheduler, **kwargs):
    scheduler.add_job(ScheduledJob(**kwargs))


@scheduler_bp.route("jobs/<job_hash>", methods=["DELETE"])
@doc(tags=["scheduler"])
@marshal_with(ScheduleResponseSchema)
@with_scheduler
def delete_job(scheduler, job_hash):
    scheduler.delete_job(job_hash)


@scheduler_bp.route("/jobs/<job_hash>", methods=["PUT"])
@doc(tags=["scheduler"])
@marshal_with(ScheduleResponseSchema)
@validate_schema(ScheduledJobSchema)
@with_scheduler
def update_job(scheduler, job_hash, **kwargs):
    scheduler.delete_job(job_hash)
    scheduler.add_job(ScheduledJob(**kwargs))


@scheduler_bp.route("/jobs", methods=["PUT"])
@doc(tags=["scheduler"])
@marshal_with(ScheduleResponseSchema)
@validate_schema(ScheduledJobsSchema)
@with_scheduler
def update_jobs(scheduler, **kwargs):
    for job_hash, job_dict in kwargs["jobs"].items():
        scheduler.delete_job(job_hash)
        scheduler.add_job(ScheduledJob(**job_dict))


scheduler_routes = [get_jobs, add_job, delete_job, update_job, update_jobs]



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from typing import TYPE_CHECKING, Any, Callable, Dict, Generic, List, Optional, Tuple, TypeVar, overload

from typing_extensions import Literal, TypedDict

from .helpers.actor_arguments import get_actor_arguments
from .logging import get_logger
from .message import Message

if TYPE_CHECKING:
    from .broker import Broker


#: The regular expression that represents valid queue names.
_queue_name_re = re.compile(r"[a-zA-Z_][a-zA-Z0-9._-]*")

F = TypeVar("F", bound=Callable[..., Any])


class ActorDict(TypedDict):
    name: str
    queue_name: str
    alternative_queues: Optional[List[str]]
    priority: int
    args: list


@overload
def actor(
    fn: Literal[None] = None,
    *,
    actor_name: Optional[str] = None,
    queue_name: str = "default",
    alternative_queues: Optional[List[str]] = None,
    priority: int = 0,
    **options: Any,
) -> "Callable[[F], Actor[F]]":
    ...


@overload
def actor(
    fn: F,
    *,
    actor_name: Optional[str] = None,
    queue_name: str = "default",
    alternative_queues: Optional[List[str]] = None,
    priority: int = 0,
    **options: Any,
) -> "Actor[F]":
    ...


def actor(
    fn: Optional[F] = None,
    *,
    actor_name: Optional[str] = None,
    queue_name: str = "default",
    alternative_queues: Optional[List[str]] = None,
    priority: int = 0,
    **options: Any,
):
    """Declare an actor.

    Examples:

      >>> import remoulade

      >>> @remoulade.actor
      ... def add(x, y):
      ...   print(x + y)
      ...
      >>> add
      Actor(<function add at 0x106c6d488>, queue_name='default', actor_name='add')

      You need to declare an actor before using it
      >>> remoulade.declare_actors([add])
      None

      >>> add(1, 2)
      3

      >>> add.send(1, 2)
      Message(
        queue_name='default',
        actor_name='add',
        args=(1, 2), kwargs={}, options={},
        message_id='e0d27b45-7900-41da-bb97-553b8a081206',
        message_timestamp=1497862448685)

    Parameters:
      fn(callable): The function to wrap.
      actor_name(str): The name of the actor.
      queue_name(str): The name of the default queue to use.
      alternative_queues(List[str]): The names of other queues this actor may be enqueued into.
      priority(int): The actor's global priority.  If two tasks have
        been pulled on a worker concurrently and one has a higher
        priority than the other then it will be processed first.
        Higher numbers represent higher priorities.
      **options(dict): Arbitrary options that vary with the set of
        middleware that you use.  See ``get_broker().actor_options``.

    Returns:
      Actor: The decorated function.
    """

    def decorator(fn: F) -> Actor[F]:
        nonlocal actor_name
        actor_name = actor_name or fn.__name__
        queues_names = [queue_name]
        if alternative_queues is not None:
            queues_names += alternative_queues
        if any(not _queue_name_re.fullmatch(name) for name in queues_names):
            raise ValueError(
                "Queue names must start with a letter or an underscore followed "
                "by any number of letters, digits, dashes or underscores."
            )

        return Actor(
            fn,
            actor_name=actor_name,
            queue_name=queue_name,
            alternative_queues=alternative_queues,
            priority=priority,
            options=options,
        )

    if fn is None:
        return decorator
    return decorator(fn)


class Actor(Generic[F]):
    """Thin wrapper around callables that stores metadata about how
    they should be executed asynchronously.  Actors are callable.

    Attributes:
      logger(Logger): The actor's logger.
      fn(callable): The underlying callable.
      broker(Broker): The broker this actor is bound to.
      actor_name(str): The actor's name.
      queue_name(str): The actor's default queue.
      alternative_queues(List[str]): The names of other queues this actor may be enqueued into.
      priority(int): The actor's priority.
      options(dict): Arbitrary options that are passed to the broker
        and middleware.
    """

    def __init__(
        self,
        fn: F,
        *,
        actor_name: str,
        queue_name: str,
        alternative_queues: Optional[List[str]],
        priority: int,
        options: Any,
    ) -> None:
        self.logger = get_logger(fn.__module__, actor_name)
        self.fn = fn
        self.broker: "Optional[Broker]" = None
        self.actor_name = actor_name
        self.queue_name = queue_name
        self.alternative_queues = alternative_queues
        self.priority = priority
        self.options = options

    def set_broker(self, broker: "Broker") -> None:
        invalid_options = set(self.options) - broker.actor_options
        if invalid_options:
            invalid_options_list = ", ".join(invalid_options)
            message = "The following actor options are undefined: %s. " % invalid_options_list
            message += "Did you forget to add a middleware to your Broker?"
            raise ValueError(message)
        self.broker = broker

    @property
    def queue_names(self):
        return [self.queue_name] + (self.alternative_queues or [])

    def message(self, *args: Any, **kwargs: Any) -> Message:
        """Build a message.  This method is useful if you want to
        compose actors.  See the actor composition documentation for
        details.

        Parameters:
          *args(tuple): Positional arguments to send to the actor.
          **kwargs(dict): Keyword arguments to send to the actor.

        Examples:
          >>> (add.message(1, 2) | add.message(3))
          pipeline([add(1, 2), add(3)])

        Returns:
          Message: A message that can be enqueued on a broker.
        """
        return self.message_with_options(args=args, kwargs=kwargs)

    def message_with_options(
        self,
        *,
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        queue_name: Optional[str] = None,
        **options: Any,
    ) -> Message:
        """Build a message with an arbitrary set of processing options.
        This method is useful if you want to compose actors.  See the
        actor composition documentation for details.

        Parameters:
          args(tuple): Positional arguments that are passed to the actor.
          kwargs(dict): Keyword arguments that are passed to the actor.
          queue_name(str): Name of the queue to put this message into when enqueued.
          **options(dict): Arbitrary options that are passed to the
            broker and any registered middleware.

        Returns:
          Message: A message that can be enqueued on a broker.
        """
        for middleware in self.broker.middleware:
            options = middleware.update_options_before_create_message(options, self.broker, self.actor_name)

        if queue_name is not None:
            if queue_name not in self.queue_names:
                raise ValueError(f"{queue_name} is not a valid queue name for actor {self.actor_name}.")
        else:
            queue_name = self.queue_name

        return Message(
            queue_name=queue_name,
            actor_name=self.actor_name,
            args=args or (),
            kwargs=kwargs or {},
            options=options,
        )

    def send(self, *args: Any, **kwargs: Any) -> Message:
        """Asynchronously send a message to this actor.

        Parameters:
          *args(tuple): Positional arguments to send to the actor.
          **kwargs(dict): Keyword arguments to send to the actor.

        Returns:
          Message: The enqueued message.
        """
        return self.send_with_options(args=args, kwargs=kwargs)

    def send_with_options(
        self,
        *,
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        queue_name: Optional[str] = None,
        delay: Optional[int] = None,
        **options: Any,
    ) -> Message:
        """Asynchronously send a message to this actor, along with an
        arbitrary set of processing options for the broker and
        middleware.

        Parameters:
          args(tuple): Positional arguments that are passed to the actor.
          kwargs(dict): Keyword arguments that are passed to the actor.
          queue_name(str): Name of the queue to enqueue this message into.
          delay(int): The minimum amount of time, in milliseconds, the
            message should be delayed by.
          **options(dict): Arbitrary options that are passed to the
            broker and any registered middleware.

        Returns:
          Message: The enqueued message.
        """
        if not self.broker:
            raise ValueError("No broker is set, did you forget to call set_broker ?")
        message = self.message_with_options(args=args, kwargs=kwargs, queue_name=queue_name, **options)
        return self.broker.enqueue(message, delay=delay)

    def as_dict(self) -> ActorDict:
        return {
            "name": self.actor_name,
            "queue_name": self.queue_name,
            "alternative_queues": self.alternative_queues,
            "priority": self.priority,
            "args": get_actor_arguments(self),
        }

    def __call__(self, *args, **kwargs):
        """Synchronously call this actor.

        Parameters:
          *args: Positional arguments to send to the actor.
          **kwargs: Keyword arguments to send to the actor.

        Returns:
          Whatever the underlying function backing this actor returns.
        """
        return self.fn(*args, **kwargs)

    def __repr__(self) -> str:
        return f"Actor({repr(self.fn)}, queue_name={repr(self.queue_name)}, actor_name={repr(self.actor_name)})"

    def __str__(self) -> str:
        return f"Actor({self.actor_name})"



import threading

from ..logging import get_logger
from .middleware import Middleware


class MaxTasks(Middleware):
    """Middleware that stop a worker if its amount of tasks completed reach max_tasks.
    If a task causes a worker to reach this limit, all other worker threads will complete their task, and the worker
    will be stopped afterwards.

    Parameters:
      max_tasks(int): The amount of tasks to complete before stopping the worker
    """

    def __init__(self, *, max_tasks: int):
        self.logger = get_logger(__name__, type(self))
        self.max_tasks = max_tasks
        self.tasks_count = 0
        self.lock = threading.Lock()

    def after_worker_thread_process_message(self, broker, thread):
        with self.lock:
            self.tasks_count += 1
            if self.tasks_count == self.max_tasks:
                self.logger.info(
                    f"Stopping worker thread as completed tasks ({self.tasks_count}) > max_tasks ({self.max_tasks})"
                )
                thread.stop()



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from ..broker import Broker, MessageProxy
from ..common import current_millis
from ..message import Message
from ..middleware import SkipMessage
from ..results import Results
from ..results.backends import LocalBackend


class LocalBroker(Broker):
    """Broker that calculate the message result immediately

    It can only be used with LocalBackend as a result backend
    """

    def __init__(self, middleware=None):
        super().__init__(middleware)

    @property
    def local(self):
        return True

    def add_middleware(self, middleware, *, before=None, after=None):
        if isinstance(middleware, Results) and not isinstance(middleware.backend, LocalBackend):
            raise RuntimeError("LocalBroker can only be used with LocalBackend.")
        super().add_middleware(middleware)

    def emit_before(self, signal, *args, **kwargs):
        # A local broker should not catch any exception because we are not in a worker but in the main thread
        for middleware in self.middleware:
            getattr(middleware, "before_" + signal)(self, *args, **kwargs)

    def emit_after(self, signal, *args, **kwargs):
        for middleware in reversed(self.middleware):
            getattr(middleware, "after_" + signal)(self, *args, **kwargs)

    def consume(self, queue_name, prefetch=1, timeout=100):
        raise ValueError("LocalBroker is not destined to use with a Worker")

    def declare_queue(self, queue_name):
        self.emit_before("declare_queue", queue_name)
        self.queues[queue_name] = None
        self.emit_after("declare_queue", queue_name)

    def _apply_delay(self, message, delay: Optional[int] = None):
        if delay is not None:
            message_eta = current_millis() + delay
            message = message.copy(options={"eta": message_eta})
        return message

    def enqueue(self, message: "Message", *, delay: Optional[int] = None) -> "Message":  # pragma: no cover
        """Enqueue a message on this broker.

        Parameters:
          message(Message): The message to enqueue.
          delay(int): The number of milliseconds to delay the message for.

        Returns:
          Message: Either the original message or a copy of it.
        """

        message = self._apply_delay(message, delay)
        self.emit_before("enqueue", message, delay)
        self.emit_after("enqueue", message, delay)

        message = self._enqueue(message, delay=delay)
        return message

    def _enqueue(self, message, *, delay=None):
        """Enqueue and compute a message.

        Parameters:
          message(Message): The message to enqueue
          delay(int): ignored
        """
        actor = self.get_actor(message.actor_name)
        message_proxy = MessageProxy(message)
        try:
            self.emit_before("process_message", message_proxy)

            res = None
            if not message_proxy.failed:
                res = actor(*message_proxy.args, **message_proxy.kwargs)

            self.emit_after("process_message", message_proxy, result=res)
        except SkipMessage:
            self.emit_after("skip_message", message)

        except BaseException as e:
            self.emit_after("process_message", message_proxy, exception=e)
            raise
        finally:
            if message_proxy.failed:
                self.emit_before("nack", message)
                self.emit_after("nack", message)
            else:
                self.emit_before("ack", message)
                self.emit_after("ack", message)

        return message_proxy

    def flush(self, _):
        pass

    def flush_all(self):
        pass

    def join(self, *_):
        return
